#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true
#
# This is a replacement for the standard hiera lookup method `yaml_data`. It has the same functionality
# except the added function of including a yaml file.
#
# ```yaml
# a:10
# include 'my_include.yaml'
#```
#
# This will add the contents of the file `my_include.yaml` into the current content. You can also
# include multiple files.
#
# The hiera backend also supports basic interpollation. You can use basic variable interpollation. You can however
# not use lookups.
# 
# The backend doesn't support nested includes.
#
require 'yaml'
Puppet::Functions.create_function('easy_type::yaml_with_include') do

  dispatch :yaml_with_include do
    param 'Hash', :options
    param 'Any', :context
  end
  
  argument_mismatch :missing_path do
    param 'Hash', :options
    param 'Puppet::LookupContext', :context
  end

  #
  # Process a yaml file that might contain include statements
  #
  def yaml_with_include(options, context)
    path = options['path']
    include_path = File.dirname(path)
    context.cached_file_data(path) do |content|
      replace_include_with_content(path, content, include_path, context) if contains_include?(content)
      begin
        data = Puppet::Util::Yaml.safe_load(content, [Symbol], path)
        if data.is_a?(Hash)
          Puppet::Pops::Lookup::HieraConfig.symkeys_to_string(data)
        else
          msg = _("%{path}: file does not contain a valid yaml hash" % { path: path })
          raise Puppet::DataBinding::LookupError, msg if Puppet[:strict] == :error && data != false
          Puppet.warning(msg)
          {}
        end
      rescue Puppet::Util::Yaml::YamlLoadError => ex
        # YamlLoadErrors include the absolute path to the file, so no need to add that
        raise Puppet::DataBinding::LookupError, _("Unable to parse %{message}") % { message: ex.message }
      end
    end
  end

  #
  # Regular expression that identifies a include statement
  # 
  def include_spec
    /^include '(\S+)'.*$/.freeze
  end

  #
  # Return true if the content contains any include statements
  #
  def contains_include?(content)
    content =~ include_spec
  end

  #
  # Return an array of the files that are included in the content
  #
  def included_files(content)
    content.scan(include_spec).flatten
  end

  #
  # Replace the content that contains any include statements, witht the
  # include statment replaced with the actual include content
  #
  def replace_include_with_content(path, content, include_path, context)
    included_files(content).each do | file|
      interpolated_file = context.interpolate(file)
      context.explain { "Interpolated file '#{file}' to '#{interpolated_file}'"} if file != interpolated_file
      context.explain { "Including file '#{include_path}/#{file}' into contents of '#{path}'" } 
      content.gsub!(/^include '#{file}'.*$/, include_contents(file, include_path, context) )
    end
  end

  #
  # Return the content of the specified include file
  #
  def include_contents(include_file, include_path, context)
    full_path = "#{include_path}/#{include_file}"
    if File.exists?(full_path) && File.file?(full_path)
      data = context.cached_file_data(full_path)
      data.gsub!(/\A---/,'')    # Remove start of yaml if it is specified
      data.gsub!(/^...\z/,'')   # Remove end of yaml if it is specified
      data
    else
      context.explain { "Included file '#{include_path}/#{include_file}' not found or not a file." }
      ''  # Replace the include with empty content so the hiera processing can continue
    end
  end

  def missing_path(options, context)
    "one of 'path', 'paths' 'glob', 'globs' or 'mapped_paths' must be declared in hiera.yaml when using this data_hash function"
  end
end   
